from .. import Module, Log, Settings
from .. import ConfigFileNotFound
from .. import CommandMessage
from ..httpclient import HTTPClient, WebResponse

from datetime import datetime

log = Log("CORE:Uptime", verbose="log")

class Uptime(Module):

    def __init__(self, channel):
        Module.__init__(self, "Uptime")
        self.channel = channel
    
    def on_command(self, data: CommandMessage):
        if "uptime" == data.KEYWORD:
            client = HTTPClient("Uptime Command Handler")
            streamURL = f"https://api.twitch.tv/helix/streams?user_login={data.channel}"
            response = client.GetFromTwitch(streamURL)
            message = ""
            if len(response.json["data"]) > 0:
                delta = datetime.utcnow() - datetime.strptime(response.json["data"][0]["started_at"],"%Y-%m-%dT%H:%M:%SZ")
                seconds = int(round(delta.total_seconds()))
                minutes, seconds = divmod(seconds, 60)
                hours, minutes = divmod(minutes, 60)
                message = "{:d} hours, {:02d} minutes".format(hours, minutes)
            else:
                message = f"{data.channel} isn't live right now {data.display_name}, please try again later when the stream is live."
            
            self.channel.socket.send_message(data.channel, message)


def setup(channel):
    if channel.channel_name.lower() == "lowkotv":
        return
    try:
        settings = Settings()
        try:
            if settings.get_setting("commands.uptime") == True:
                channel.load_module(Uptime(channel))
                log.log(f"[{channel.channel_name}]: Loaded Module Uptime")
        except KeyError:
            channel.load_module(Uptime(channel))
            log.log(f"[{channel.channel_name}]: Loaded Module Uptime")
    except ConfigFileNotFound:
        channel.load_module(Uptime(channel))
        log.log(f"[{channel.channel_name}]: Loaded Module Uptime")

def teardown(channel):
    log.log(f"[{channel.channel_name}]: Removed Module Uptime")


from .model import Model
from .helpers import Helpers
from .settings import Settings
from pathlib import Path
from .errors import MissingSetting, ConfigFileNotFound
from pymysql.err import OperationalError

class _bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class Log():
    log_format: str
    instance: str
    verbose: list
    client: str
    model: Model

    def __init__(self, instance, client="Twitch", verbose=None):
        self.enable_db = False
        self.verbose = ["debug", "log", "info", "warn", "error"]
        self.log_format = "\033[92m{time}\033[0m ({level}) [{instance}]: {message}"
        self.chat_log_format = "\033[92m{time}\033[0m ({level}) [{instance}] {chatter}: {message}"
        self.log_file_format = "{time} ({level}) [{instance}]: {message}\n"
        self.instance = instance
        self.client = client
        try:
            settings = Settings()
            db_user = settings.get_setting("database.user")
            db_pass = settings.get_setting("database.pass")
            self.enable_db = True
            try:
                self.model = Model(user=db_user, password=db_pass)
                self.__confirm_log_table()
            except OperationalError:
                self.enable_db = False
                self.debug("Database connection could not be established, ensure the databse is available.")
        except ConfigFileNotFound:
            # Config file doesn't exist
            pass
        except KeyError:
            # database setting key does not exist
            pass
        except MissingSetting:
            # Databse Settings don't exist
            pass

        if verbose is None:
            try:
                verbose = Settings().get_setting("log")
            except MissingSetting:
                verbose = "log"
            except ConfigFileNotFound:
                verbose = "log"
        

        if verbose == "debug":
            self.verbose = ["debug", "log", "info", "warn", "error"]
        elif verbose == "log":
            self.verbose = ["log", "info", "warn", "error"]
        elif verbose == "info":
            self.verbose = ["info", "warn", "error"]
        elif verbose == "warn":
            self.verbose = ["warn", "error"]
        elif verbose == "error":
            self.verbose = ["error"]


    def debug(self, message):
        if "debug" in self.verbose:
            self.__render({
                "time": Helpers().get_timestamp(),
                "level": "dbug",
                "instance": self.instance,
                "message": message
            })
        self.__record(message, "debug")

    def log(self, message):
        if "log" in self.verbose:
            self.__render({
                "time": Helpers().get_timestamp(),
                "level": "log_",
                "instance": self.instance,
                "message": message
            })
        self.__record(message, "log")

    
    def info(self, message):
        if "info" in self.verbose:        
            self.__render({
                "time": Helpers().get_timestamp(),
                "level": f"{_bcolors.OKBLUE}info{_bcolors.ENDC}",
                "instance": self.instance,
                "message": message
            })
        self.__record(message, "info")

    def sent(self, message, channel, jid=None, sid=None):
        if "info" in self.verbose:        
            self.__render_chat({
                "time": Helpers().get_timestamp(),
                "level": f"{_bcolors.HEADER}sent{_bcolors.ENDC}",
                "instance": f"{_bcolors.OKGREEN}{channel}{_bcolors.ENDC}",
                "chatter": f"{_bcolors.HEADER}BOT{_bcolors.ENDC}",
                "message": message
            })
        self.__record(f"[BOT::Sent]: {message}", "chat")
        if jid is not None:
            self.__record_chat(message, "BOT::Sent", channel, jid, sid)
        
    
    def sent_whisper(self, message, channel):
        if "info" in self.verbose:        
            self.__render_chat({
                "time": Helpers().get_timestamp(),
                "level": f"{_bcolors.HEADER}whisper{_bcolors.ENDC}",
                "instance": f"{_bcolors.OKGREEN}{channel}{_bcolors.ENDC}",
                "chatter": f"{_bcolors.HEADER}BOT{_bcolors.ENDC}",
                "message": message
            })
        self.__record(f"[BOT::Sent]: {message}", "whisper")
    
    def whisper(self, message, channel):
        if "info" in self.verbose:        
            self.__render({
                "time": Helpers().get_timestamp(),
                "level": f"{_bcolors.OKBLUE}whisper{_bcolors.ENDC}",
                "instance": f"{_bcolors.OKGREEN}{channel}{_bcolors.ENDC}",
                "message": message
            })
        self.__record(message, "whisper")
    
    def chat(self, message, channel, chatter, jid=None, sid=None):
        if "info" in self.verbose:        
            self.__render_chat({
                "time": Helpers().get_timestamp(),
                "level": f"{_bcolors.OKBLUE}chat{_bcolors.ENDC}",
                "instance": f"{_bcolors.OKGREEN}{channel}{_bcolors.ENDC}",
                "chatter": f"{_bcolors.WARNING}{chatter}{_bcolors.ENDC}",
                "message": message
            })
        self.__record(f"[{channel}].[{chatter}]: {message}", "chat")
        if jid is not None:
            self.__record_chat(message, chatter, channel, jid, sid)

    
    def warn(self, message):
        if "warn" in self.verbose:
            self.__render({
                "time": Helpers().get_timestamp(),
                "level": f"{_bcolors.WARNING}warn{_bcolors.ENDC}",
                "instance": self.instance,
                "message": f"{_bcolors.WARNING}{message}{_bcolors.ENDC}"
            })
        self.__record(message, "warn")
        self.__send_pushover_alert(message, "warn")

    
    def error(self, message):
        if "error" in self.verbose:
            self.__render({
                "time": Helpers().get_timestamp(),
                "level": f"{_bcolors.FAIL}errr{_bcolors.ENDC}",
                "instance": self.instance,
                "message": f"{_bcolors.FAIL}{message}{_bcolors.ENDC}"
            })
        self.__record(message, "error")
        self.__send_pushover_alert(message, "error")

    def exception(self, message):
        if "error" in self.verbose:
            self.__render({
                "time": Helpers().get_timestamp(),
                "level": f"{_bcolors.FAIL}excp{_bcolors.ENDC}",
                "instance": self.instance,
                "message": f"{_bcolors.FAIL}{message}{_bcolors.ENDC}"
            })
        self.__record(message, "exception")
        self.__send_pushover_alert(message, "exception")

    def set_verbose(self, verbose):
        self.info(f"Updating verbosity to '{verbose}'")
        if verbose == "debug":
            self.verbose = ["debug", "log", "info", "warn", "error"]
        elif verbose == "log":
            self.verbose = ["log", "info", "warn", "error"]
        elif verbose == "info":
            self.verbose = ["info", "warn", "error"]
        elif verbose == "warn":
            self.verbose = ["warn", "error"]
        elif verbose == "error":
            self.verbose = ["error"]

    def set_instance(self, instance):
        self.instance = instance

    def __render(self, data):
        print(self.log_format.format(**data))
    
    def __render_chat(self, data):
        print(self.chat_log_format.format(**data))


    def __record_chat(self, message, chatter, channel, jarvis_id, stream_id = None, retry = True):
        sanitisedMessage = message \
            .replace("'", "''") \
            .replace("\"","\"\"")
        if stream_id is None: 
            query = f"""
            INSERT INTO Channel_Chat
            (IdentityID, Channel, Chatter_DisplayName, Message)
            VALUES
            ('{jarvis_id}','{channel}','{chatter}','{sanitisedMessage} ')
            """
        else:
            query = f"""
            INSERT INTO Channel_Chat
            (IdentityID, SRID, Channel, Chatter_DisplayName, Message)
            VALUES
            ('{jarvis_id}','{stream_id}','{channel}','{chatter}','{sanitisedMessage} ')
            """
        try:
            if self.enable_db:
                self.model.insert(query)
        except Exception as ex:
            print(f"{_bcolors.FAIL}An exception occurred saving the log to the database{_bcolors.ENDC}")
            print(f"Exception: {ex}")
            print(f"Query: {query}")
            self.__heal_model()
            if retry:
                print("Attempting single retry of db save")
                self.__record_chat(message, chatter, channel, jarvis_id, stream_id, False)
            else:
                print("retried attempt failed - couldn't save log to db")

    def __record(self, message, level, retry = True):
        sanitisedMessage = message \
            .replace("'", "''") \
            .replace("\"","\"\"")
        query = f"insert into Logs (Source, Message, LogLevel, LogTime, Component) values ('{self.instance}', '{sanitisedMessage} ', '{level}', '{Helpers().get_timestamp()}', '{self.client}')"
        try:
            if self.enable_db:
                self.model.insert(query)
        except Exception as ex:
            print(f"{_bcolors.FAIL}An exception occurred saving the log to the database{_bcolors.ENDC}")
            print(f"Exception: {ex}")
            print(f"Query: {query}")
            self.__heal_model()
            if retry:
                print("Attempting single retry of db save")
                self.__record(message, level, False)
            else:
                print("retried attempt failed - couldn't save log to db")
    
    def __send_pushover_alert(self, message: str, log_level: str):
        pass

    
    def __heal_model(self):
        print("Attempting to heal Model Object")
        settings = Settings()
        db_user = settings.get_setting("database.user")
        db_pass = settings.get_setting("database.pass")
        self.model = Model(user=db_user, password=db_pass)

    
    def __confirm_log_table(self):
        query = "SHOW TABLES LIKE 'Logs';"
        logsExist = self.model.fetchOne(query)
        if logsExist is None:
            create_logs = """CREATE TABLE `Logs` (
                `ID` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
                `Source` VARCHAR(255) NOT NULL,
                `Message` TEXT NOT NULL,
                `LogLevel` VARCHAR(100) NOT NULL DEFAULT 'Undefined',
                `LogTime` DATETIME NOT NULL DEFAULT CURRENT_TIMESTAMP,
                `Client` VARCHAR(255) NULL DEFAULT NULL,
                `Component` VARCHAR(255) NULL DEFAULT NULL,
                PRIMARY KEY (`ID`)
            )
            COLLATE='utf8mb4_bin'
            ENGINE=InnoDB
            ;"""
            self.model.update(create_logs)
            if self.model.fetchOne(query) is None:
                self.enable_db = False
                self.error("Failed to create Logs table")
            else:
                self.log("Logs table created successfully")






from requests import get, patch, put, post
from .settings import Settings
from .log import Log
from .model import Model
from .errors import DBConnectionUndefined, HTTPDriverNotConfigured
from urllib import parse

class WebResponse():
    def __init__(self, data:dict):
        self.status = data["status"]
        self.message = data["message"]
        self.code = data["code"]
        self.text = data["text"]
        self.json = data["json"]
        self.headers = data["headers"]

    def __repr__(self):
        return f"[{self.status}]({self.message}): {self.text}"


class HTTPClient():

    def __init__(self, owner:str, db_user=None, db_pass=None, twitch_oauth=None):
        self.log = Log("Web Handler", f"Web Handler: {owner}")
        self.useModel = False
        if db_user is not None and db_pass is not None:
            self.model = Model(user=db_user, password=db_pass)
            self.useModel = True
        self.twitchToken = twitch_oauth

    def GetFromTwitch(self, url: str) -> dict:
        headers = self.__get_twitch_headers(url)
        return self.Get(url, headers=headers)

    def GetFromDiscord(self, url: str) -> dict:        
        headers = self.__get_discord_headers()
        return self.Get(url, headers=headers)

    def PostFormToTwitch(self, url:str,  body: dict):
        headers = self.__get_twitch_headers(url)
        return self.PostForm(url=url, body=body, headers=headers)

    def PostFormToDiscord(self, url:str,  body: dict):
        headers = self.__get_discord_headers()
        return self.PostForm(url=url, body=body, headers=headers)

    def PostJSONToTwitch(self, url:str,  body: dict):
        headers = self.__get_twitch_headers(url)
        return self.PostJSON(url=url, body=body, headers=headers)

    def PostJSONToDiscord(self, url:str,  body: dict):
        headers = self.__get_discord_headers()
        return self.PostJSON(url=url, body=body, headers=headers)



    def Get(self, url: str, headers: dict = None) -> dict:
        """Get request"""           
        self.log.debug(f"Sending new GET request to '{url}'.\nHeaders: \n{headers}")
        try:
            response = get(url=url, headers=headers)
        except Exception as ex:
            return self.__exception_response(str(ex))
        rt = self.__process_response(response)
        self.log.debug("returning: \n{}".format(rt))
        return rt


    
    def PostForm(self, url:str,  body: dict, headers: dict = None):
        if headers is None:
            setting = Settings()
            headers = {
                "Client-ID": setting.get_setting("tokens.twitch.user"), 
                "content-type": 'application/x-www-form-urlencoded', 
            }
        self.log.debug(f"Sending new POST request to '{url}'.\nHeaders: \n{headers}\nBody:\n{body}")
        try:
            response = post(url=url, headers=headers, data=body)
        except Exception as ex:
            return self.__exception_response(str(ex))
        rt = self.__process_response(response)
        self.log.debug("returning: \n{}".format(rt))
        return rt
    
    def PostJSON(self, url:str,  body: dict, headers: dict = None):
        if headers is None:
            setting = Settings()
            headers = {
                "Client-ID": setting.get_setting("tokens.twitch.user"), 
                "content-type": 'application/json', 
            }
        self.log.debug(f"Sending new POST request with JSON body to '{url}'.\nHeaders: \n{headers}\nBody:\n{body}")
        response = post(url=url, headers=headers, json=body)
        try:
            rt = self.__process_response(response)
        except Exception as ex:
            return self.__exception_response(str(ex))
        self.log.debug("returning: \n{}".format(rt))
        return rt
    
    def __exception_response(self, message):
        self.log.warn("An exception occrred when attempting to contact the server. Exception: " + message)
        return WebResponse ({
            "status": "error",
            "message": "an exception occurred when attempting to contact server.",
            "code": 600,
            "text": message,
            "json": {"error": message},
            "headers": None
        })


    def __process_response(self, response):
        if response.status_code == 200:
            self.log.debug("Response: 200 | OK \n Headers:\n{}\n\nContent:\n{}".format(response.headers, response.text))
            return WebResponse ({
                "status": "ok",
                "message": "response ok",
                "code": response.status_code,
                "text": response.text,
                "json": response.json(),
                "headers": response.headers
            })
        elif response.status_code == 201:
            self.log.debug("Response: 201 | OK \n Headers:\n{}\n\nContent:\n{}".format(response.headers, response.text))
            return WebResponse ({
                "status": "ok",
                "message": "no data returned",
                "code": response.status_code,
                "text": "No Data",
                "json": {"No Data"},
                "headers": response.headers
            })
        elif response.status_code == 202:
            self.log.debug("Response: 202 | Accepted \n Headers:\n{}\n\nContent:\n{}".format(response.headers, response.text))
            return WebResponse ({
                "status": "ok",
                "message": "accepted",
                "code": response.status_code,
                "text": "No Data",
                "json": {"No Data"},
                "headers": response.headers
            })
        elif response.status_code == 204:
            self.log.debug("Response: 204 | No Content \n Headers:\n{}\n\nContent:\n{}".format(response.headers, response.text))
            return WebResponse ({
                "status": "ok",
                "message": "no content",
                "code": response.status_code,
                "text": "No Data",
                "json": {"No Data"},
                "headers": response.headers
            })
        elif response.status_code == 400:
            self.log.info("Response: 400 | Bad Request \n Headers:\n{}\n\nContent:\n{}".format(response.headers, response.text))
            return WebResponse ({
                "status": "error",
                "message": "bad request",
                "code": response.status_code,
                "text": response.text,
                "json": response.json(),
                "headers": response.headers
            })
        elif response.status_code == 401:
            self.log.info("Response: 401 | Unauthorised Request \n Headers:\n{}\n\nContent:\n{}".format(response.headers, response.text))
            return WebResponse ({
                "status": "error",
                "message": "unauthorised request",
                "code": response.status_code,
                "text": response.text,
                "json": response.json(),
                "headers": response.headers
            })
        elif response.status_code == 404:
            self.log.info("Response: 404 | Resource not found \n Headers:\n{}\n\nContent:\n{}".format(response.headers, response.text))
            return WebResponse ({
                "status": "error",
                "message": "resource not found",
                "code": response.status_code,
                "text": response.text,
                "json": response.json(),
                "headers": response.headers
            })
        else:
            self.log.error("Response: {} | \nHeaders: {} \nError Response\n{}".format(response.status_code, response.headers, response.text))
            return WebResponse ({               
                "status": "error",
                "message": "server returned an error",
                "code": response.status_code,
                "text": response.text,
                "json": response.json(),
                "headers": response.headers
            })

    def __get_twitch_headers(self, url: str) -> dict:
        setting = Settings()
        token = ""

        if self.twitchToken is None and not self.useModel:
            raise HTTPDriverNotConfigured("Values provided for DB connection / Twitch OAuth are invalid.\nYou must provide either:\n - Values for the Databse with 'db_user' and 'db_pass'\n - An OAuth token for the Twitch API")

        elif self.twitchToken is not None:
            token = self.twitchToken
        
        else:
            jOAUTH = self.model.fetchOne("select AccessToken from Tokens where TokenType = 'JarvisAppToken'")
            token = jOAUTH[0]

        if "kraken" in url:
            headers = {
                "Client-ID" : setting.get_setting("tokens.twitch.user"),
                "Authorization": "OAuth " + token,
                "Accept": "application/vnd.twitchtv.v5+json"
            }
        else:
            headers = {
                "Client-ID" : setting.get_setting("tokens.twitch.user"),
                "Authorization": "Bearer " + token
            }

        return headers

    
    def __get_discord_headers(self) -> dict:
        setting = Settings()
        headers = {
            "Client-ID": "{}".format(setting.get_setting("integration.discord.client")),
            "User-Agent": setting.get_setting("integration.discord.agent"),
            "Authorization": "Bot {}".format(setting.get_setting("integration.discord.key"))
        }
        return headers


import traceback
import socket
from time import sleep
from datetime import datetime


from .log import Log
# from .bridge import Bridge
from threading import Thread
from .message import RawMessage


from .messageparser import parse_line


class Socket():
    
    # bridge: Bridge
    socket: socket.socket
    buffer: str
    active: bool
    ready: bool
    channel_list: list
    log: Log
    last_ping: datetime


    # def __init__(self, bridge: Bridge = None):
    def __init__(self, channel, nick: str, token: str, id = None, verbose=False):
        self.name = f"Thread - {channel.channel_name} Socket"
        self.nick = nick
        self.token = token
        self.active = True
        self.verbose = verbose
        self.buffer = ""
        self.id = id
        self.stream_id = None
        self.socket = None
        self.channel = channel
        self.log = Log(f"{channel.channel_name} Socket")
        self.__connect()
        
        

    def __connect(self):
        self.log.log("Engageing Socket. Standby...")
        self.buffer = ""
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect(("irc.chat.twitch.tv", 6667))
        self._send_raw(f"PASS {self.token}")
        self._send_raw(f"NICK {self.nick}")
        self._send_raw("CAP REQ :twitch.tv/membership")
        self._send_raw("CAP REQ :twitch.tv/tags")
        self._send_raw("CAP REQ :twitch.tv/commands")
        self._send_raw(f"JOIN #{self.channel.channel_name} ")
        self.last_ping = datetime.now()
        self.log.info("Socket engaged.")
        pass

    def disconnect(self):
        if self.verbose:
            print("departing channels")
        try:
            self._send_raw(f"PART #{self.channel.channel_name} ")
        except Exception as e:
            self.log.exception(f"Suppressing a caught an exception in `Socket.disconnect()` [Parting channel]. Details below\n{type(e)}: {traceback.format_exc()}")
        try:
            self.socket.close()
        except Exception as e:
            self.log.exception(f"Suppressing a caught an exception in `Socket.disconnect()` [closing socket]. Details below\n{type(e)}: {traceback.format_exc()}")



    def reconnect(self):
        if self.verbose:
            print("Reconnect detected!")
        self.disconnect()
        if self.verbose:
            print("Waiting to reconnect.")
        sleep(10)
        self.__connect()

    def connect_to_channel(self, channel: str):
        self._send_raw("JOIN #"+ channel.lower() +" ")
        # self.channel_list.append(channel)
        if self.verbose:
            print(f"connecting to '{channel}'")

    def disconnect_from_channel(self, channel: str):
        self._send_raw("PART #"+ channel.lower() +" ")
        # self.__clearchannel(channel)
        if self.verbose:
            print(f"departing from '{channel}'")

    def __clearchannel(self, channel: str):
        counter = 0
        for chn in self.channel_list:
            if chn.name == channel:
                self.channel_list.pop(counter)
                return
            counter += 1

    def listenToStream(self):
        try:
            self.__process_stream_data()
        except Exception as e:
            self.log.exception(f"Suppressing a caught an exception in `Socket.listenToStream()`, will attempt to reconnect to recover the connection and continue without raising. Details below\n{type(e)}: {traceback.format_exc()}")
            self.reconnect()
            self.log.info("Reconnect attempt executed.")

    def run(self):
        while self.active:
            self.__process_stream_data()
        try:
            self.socket.close()
        except Exception as e:
            self.log.exception(f"Suppressing a caught an exception while attempting to close the socket in `Socket.run()`. Details below\n{type(e)}: {traceback.format_exc()}")

    def close(self):
        if self.verbose:
            print(f"({self.name}) Closing Socket")
        self.active = False
        self.socket.close()
        


    def _send_raw(self, message: str):
        try:
            self.socket.send((f"{message}\r\n").encode('utf-8'))
            if self.verbose:
                if message[:4] == "PASS":
                    print(f"({self.name}) < PASS ****")
                else:
                    print(f"({self.name}) < {message}")
        except OSError:
            self.log.error(f"Socket is closed and must be reopened to send the message '{message}'")


    def __process_stream_data(self):
        try:
            self.buffer = self.buffer + self.socket.recv(1024).decode()
        except ConnectionAbortedError:
            self.log.info("Socket connection has Closed")
        except UnicodeDecodeError:
            self.log.warn(f"Unicode Decode error detected, possible issue with the buffer.\nBuffer: [{self.buffer}]\n\nRegenerating buffer...")
            self.buffer = ""
            self.log.info("Buffer regeneration completed.")
        except OSError:
            self.log.warn("OSError detected, socket issue identitfied. Attempting to recover socket.")
            self.reconnect()

        temp = self.buffer.split("\n")
        self.buffer = temp.pop()
        for line in temp:
            if ("PING :tmi.twitch.tv" in line): # Keep Alive Mechanism
                self._send_raw("PONG :tmi.twitch.tv")
                self.last_ping = datetime.now()
            self.on_socket_data(line)

    def health_check(self):
        if (datetime.now() - self.last_ping).seconds > 420:
            self.log.error(f"extended delay detected between pings, manually calling reconnect.[PING timeout: {(datetime.now() - self.last_ping).seconds} sec]")
            self.reconnect()

    
    def set_stream_id(self, new_id):
        self.stream_id = new_id


    def on_socket_data(self, line: str):
        self.on_data(parse_line(line))
        
    def on_data(self, data):
        if self.verbose:
            print(f" >", data.line)
        self.channel.on_raw(data)
        
        if data.inner == "Message":
            self.channel.on_message(data)
        elif data.inner == "Join":
            self.channel.on_join(data)
        elif data.inner == "Mode":
            self.channel.on_mode(data)
        elif data.inner == "Names":
            self.channel.on_names(data)
        elif data.inner == "Part":
            self.channel.on_part(data)
        elif data.inner == "ClearChat":
            self.channel.on_clearchat(data)
        elif data.inner == "ClearMessage":
            self.channel.on_clearmessage(data)
        elif data.inner == "HostTarget":
            self.channel.on_hosttarget(data)
        elif data.inner == "Notice":
            self.channel.on_notice(data)
        elif data.inner == "Reconnect":
            self.channel.on_reconnect(data)
        elif data.inner == "RoomState":
            self.channel.on_roomstate(data)
        elif data.inner == "UserState":
            self.channel.on_userstate(data)
        elif data.inner == "GlobalUserState":
            self.channel.on_globaluserstate(data)
        elif data.inner == "UserNotice":
            self.channel.on_usernotice(data)
        elif data.inner == "RitualUserNotice":
            self.channel.on_ritual_usernotice(data)
        elif data.inner == "BitBadgeUpgradeUserNotice":
            self.channel.on_bitbadgeupgrade_usernotice(data)
        elif data.inner == "RaidUserNotice":
            self.channel.on_raid_usernotice(data)
        elif data.inner == "Whisper":
            if data.channel == self.channel.channel_name:
                self.log.whisper (data.message_text, data.display_name)
                self.channel.on_whisper(data)
        elif data.inner == "SubscriberUserNotice":
            if data.display_name.lower() != self.nick.lower():
                self.channel.on_subscriber_usernotice(data)
        elif data.inner == "GiftedSubscriberUserNotice":
            if data.display_name.lower() != self.nick.lower():
                self.channel.on_giftedsubscriber_usernotice(data)
        elif data.inner == "PrivateMessage":
            if data.display_name.lower() != self.nick.lower():
                self.log.chat(data.message_text, data.channel, data.display_name, jid=self.id, sid=self.stream_id)
                self.channel.on_privmessage(data)
        elif data.inner == "CommandMessage":
            if data.display_name.lower() != self.nick.lower():
                self.log.chat(f"[CMD] {data.message_text}",data.channel, data.display_name, jid=self.id, sid=self.stream_id)
                self.channel.on_command(data)

    def join(self, channel: str):
        send = f"JOIN #{channel.lower()}"
        self._send_raw(send)

    def send_message(self, channel: str, message: str):
        send = f"PRIVMSG #{channel.lower()} :{message}"
        self._send_raw(send)
        self.log.sent(message, channel.lower(), jid=self.id, sid=self.stream_id)
    
    def send_action_message(self, channel: str, message: str):
        send = f"PRIVMSG #{channel.lower()} :/me {message}"
        self._send_raw(send)
        self.log.sent(f"ACTION: {message}", channel.lower(), jid=self.id, sid=self.stream_id)

    def send_whisper(self, user: str, message: str):
        # send = f"PRIVMSG #{user.lower()} :/w {user.lower()} {message}"
        send = f"PRIVMSG #{self.nick.lower()} :/w {user.lower()} {message}"
        self._send_raw(send)
        self.log.sent_whisper(message, user.lower())

    def timeout_user(self, user: str, channel: str, timeout=1):
        send = f"PRIVMSG #{channel} :/timeout {user} {timeout}"
        self._send_raw(send)
        self.log.sent(f"TIMEOUT: Timed out user {user} for {timeout} seconds in {channel}'s stream chat", channel.lower(), jid=self.id, sid=self.stream_id)
        self.log.info(f"TIMEOUT: Timed out user {user} for {timeout} seconds in {channel}'s stream chat")

    def clear_message(self, channel: str, message_id: str):
        send = f"PRIVMSG #{channel} :/delete {message_id}"
        self._send_raw(send)
        self.log.sent(f"CLEAR MESSAGE: Deleted a message (ID: {message_id}) from {channel}'s stream chat", channel.lower(), jid=self.id, sid=self.stream_id)
        self.log.info(f"CLEAR MESSAGE: Deleted a message (ID: {message_id}) from {channel}'s stream chat")

    def ban_user(self, user: str, channel: str):
        send = f"PRIVMSG #{channel} :/ban {user}"
        self._send_raw(send)
        self.log.sent(f"BAN: Banned user {user} from {channel}'s stream chat", channel.lower(), jid=self.id, sid=self.stream_id)
        self.log.info(f"BAN: Banned user {user} from {channel}'s stream chat")
